# Copyright (c) 2019 Santosh Philip
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
"""the schema used for testing is in this module
It will be loaded only once, since that is how imports work"""

import os

from eppy3000 import epschema

THIS_DIR = os.path.dirname(os.path.abspath(__file__))

RESOURCES_DIR = os.path.join(THIS_DIR, os.pardir, "eppy3000", "resources")
VERSION = "9-0-1"  # current default for integration tests on local system


SCHEMA_FILES = os.path.join(RESOURCES_DIR, "schema")
TEST_SCHEMA = f"V{VERSION[:3].replace('-', '_')}/Energy+.schema.epJSON"
schema_file = os.path.join(SCHEMA_FILES, TEST_SCHEMA)

schema = epschema.read_epschema_asmunch(schema_file)


# IDD_FILES = os.path.join(RESOURCES_DIR, 'iddfiles')
# IDF_FILES = os.path.join(RESOURCES_DIR, 'idffiles')
# try:
#     VERSION = os.environ["ENERGYPLUS_INSTALL_VERSION"]  # used in CI files
# except KeyError:
#     VERSION = '8-9-0'
#        # current default for integration tests on local system
# TEST_IDF = "V{}/smallfile.idf".format(VERSION[:3].replace('-', '_'))
# TEST_EPW = 'USA_CA_San.Francisco.Intl.AP.724940_TMY3.epw'
# TEST_IDD = "Energy+V{}.idd".format(VERSION.replace('-', '_'))
# TEST_OLD_IDD = 'Energy+V7_2_0.idd'
#
#
# @pytest.fixture()
# def test_idf():
#     idd_file = os.path.join(IDD_FILES, TEST_IDD)
#     idf_file = os.path.join(IDF_FILES, TEST_IDF)
#     modeleditor.IDF.setiddname(idd_file, testing=True)
#     idf = modeleditor.IDF(idf_file, TEST_EPW)
#     try:
#         ep_version = idf.idd_version
#         assert ep_version == versiontuple(VERSION)
#     except AttributeError:
#         raise
#     return idf
#



# Copyright (c) 2018 Santosh Philip
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
"""py.test for readepj.py"""

import json
from io import StringIO
import eppy3000.readepj as readepj
from eppy3000.epMunch import EPMunch


def test_readepjjson():
    """py.test for readepjjson"""
    dct = dict(
        a=dict(aa=dict(z=-1, y=-2), bb=dict(zz=-11, yy=-22)),
        b=dict(cc=dict(ab=12, bc=23), dd=dict(cd=34, de=45)),
    )
    dctstr = json.dumps(dct)
    fhandle = StringIO(dctstr)
    result = readepj.readepjjson(fhandle)
    assert isinstance(result, EPMunch)


def test_addeppykeys():
    """py.test for addeppykeys"""
    dct = dict(
        a=dict(aa=dict(z=-1, y=-2), bb=dict(zz=-11, yy=-22)),
        b=dict(cc=dict(ab=12, bc=23), dd=dict(cd=34, de=45)),
    )
    dctstr = json.dumps(dct)
    fhandle = StringIO(dctstr)
    epmunch = readepj.readepjjson(fhandle)
    epobject = epmunch.a.aa
    assert "eppykey" in epobject
    assert "eppyname" in epobject


def test_removeeppykeys():
    """py.test for removeeppykeys"""
    dct = dict(
        a=dict(aa=dict(z=-1, y=-2), bb=dict(zz=-11, yy=-22)),
        b=dict(cc=dict(ab=12, bc=23), dd=dict(cd=34, de=45)),
    )
    dctstr = json.dumps(dct)
    # test for rkeys=None
    fhandle = StringIO(dctstr)
    epmunch = readepj.readepjjson(fhandle)
    readepj.removeeppykeys(epmunch)
    epobject = epmunch.a.aa
    assert "eppykey" not in epobject
    assert "eppyname" not in epobject
    # test for rkeys=['eppykey']
    fhandle = StringIO(dctstr)
    epmunch = readepj.readepjjson(fhandle)
    readepj.removeeppykeys(epmunch, rkeys=["eppykey"])
    epobject = epmunch.a.aa
    assert "eppykey" not in epobject
    assert "eppyname" in epobject



# Copyright (c) 2018-2019 Santosh Philip
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
"""read epj json file and have eppy like functionality"""

import json

# one works from ipython, the other from the script
# try:
#     from eppy3000.epMunch import EPMunch
# except ModuleNotFoundError as e:
#     from epMunch import EPMunch
from eppy3000.epMunch import EPMunch


def readepjjson(fhandle):
    """read an json epj

    Parameters
    ----------
    fhandle: io.String, io.TextIOWrapper
        can be a file open for read or a io.StringIO object

    Returns
    -------
    eppy.EPMunch
    """
    try:
        fhandle = open(fhandle, "r")
    except TypeError as e:
        pass
    as_json = json.load(fhandle)
    as_munch = EPMunch.fromDict(as_json)
    addeppykeys(as_munch)
    return as_munch


def addeppykeys(epmunch):
    """adds eppykeys needed by eppy3000

    The way E+ json is structured:

    - the lowest branch on the json tree is the epobject
    - once you are in the epobject:
        - the epobject will not know:
            1. Its own name
            2. the type of object it is
        - it will know only its own fields

    We add two new keys to the epobject that will
    let it store its name and object type
    These names start with `eppy` to make it clear that
    eppy300 has added them

    - eppykey -> stores the E+ object type (epkey)
    - eppyname -> stores the name of the object. its key in E+ json

    Parameters
    ----------
    epmunch: eppy3000.epMunch.EPMunch
        This is the E+ file as seen by eppy3000

    Returns
    -------
    None

    """
    for key, epobjects in epmunch.items():
        for name, epobject in epobjects.items():
            epobject["eppykey"] = key
            epobject["eppyname"] = name
            epobject["eppy_epobjects"] = epobjects


def removeeppykeys(epmunch, rkeys=None):
    """remove the eppykeys

    This will remove all the additional keys that eppy added
    in addeppykeys(). This is usually called before saving

    Parameters
    ----------
    epmunch: eppy3000.epMunch.EPMunch
        This is the E+ file as seen by eppy3000
    rkeys: list
        These are the keys to be removed. if rkeys is None then
        rkeys is set to ['eppykey', 'eppyname', 'eppy_objepschema']

    Returns
    -------
    None
    """
    if not rkeys:
        rkeys = [
            "eppykey",
            "eppyname",
            "eppy_objepschema",
            "eppy_epj",
            "eppy_epobjects",
        ]
    for key, epobjects in epmunch.items():
        for name, epobject in epobjects.items():
            for rkey in rkeys:
                epobject.pop(rkey, None)



# Copyright (c) 2021 Santosh Philip
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================

"""py.test for installlocation"""

import pytest
from eppy3000 import installlocation


@pytest.mark.parametrize(
    "version, expected",
    [
        ("9.0.0", "9-0-0"),  # version, expected
        ("9.0.1", "9-0-1"),  # version, expected
        ("9.2.3", "9-2-3"),  # version, expected
        ("9.2", "9-2-0"),  # version, expected
        ("9", "9-0-0"),  # version, expected
    ],
)
def test_version2folder(version, expected):
    """py.test for version2folder"""
    result = installlocation.version2folder(version)
    assert result == expected


@pytest.mark.parametrize(
    "version, platform_system, expected",
    [
        (
            "9.3",
            "Darwin",
            "/Applications/EnergyPlus-9-3-0",
        ),  # version, platform_system, expected
        ("9.3", "Windows", "C:/EnergyPlusV9-3-0"),  # version, platform_system, expected
        (
            "9.3",
            "Linux",
            "/usr/local/EnergyPlus-9-3-0",
        ),  # version, platform_system, expected
    ],
)
def test_installfolder(version, platform_system, expected):
    """py.test for installfolder"""
    result = installlocation.installfolder(version, platform_system)
    assert result == expected


@pytest.mark.parametrize(
    "version, platform_system, expected",
    [
        (
            "9.3",
            "Darwin",
            "/Applications/EnergyPlus-9-3-0/Energy+.schema.epJSON",
        ),  # version, platform_system, expected
        (
            "9.3",
            "Windows",
            "C:/EnergyPlusV9-3-0/Energy+.schema.epJSON",
        ),  # version, platform_system, expected
        (
            "9.3",
            "Linux",
            "/usr/local/EnergyPlus-9-3-0/Energy+.schema.epJSON",
        ),  # version, platform_system, expected
    ],
)
def test_schemapath(version, platform_system, expected):
    """py.test for schemapath"""
    result = installlocation.schemapath(version, platform_system)
    assert result == expected



# Copyright (c) 2021 Santosh Philip
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
"""HTML viewer for EPJ and epMunch"""


import tempfile

try:
    from IPython.display import IFrame
except ModuleNotFoundError as e:
    pass
from munch import Munch
import json2html


class JupyterNotInstalled(Exception):
    pass


def removeeppykeys_inepmunch(epmunch, rkeys=None):
    """remove the eppy keys in epmunch"""
    if not rkeys:
        rkeys = [
            "eppykey",
            "eppyname",
            "eppy_objepschema",
            "eppy_epj",
            "eppy_epobjects",
        ]
    for rkey in rkeys:
        epmunch.pop(rkey, None)


def _epmunch2displaymunch(epmunch):
    """modify epmunch for display in HTML"""
    todict = epmunch.toDict()
    todict = Munch.fromDict(todict)
    eppykey = todict["eppykey"]
    eppyname = todict["eppyname"]
    removeeppykeys_inepmunch(todict)
    en = Munch()
    en[eppyname] = todict
    ek = Munch()
    ek[eppykey] = en
    return ek


def epmunch2dct(epmunch):
    """convert epmunch to and json string"""
    ek = _epmunch2displaymunch(epmunch)
    dct = ek.toDict()
    return dct


def epmunch2html(epmunch):
    """convert epmunch to an HTML table"""
    dct = epmunch2dct(epmunch)
    html = json2html.json2html.convert(json=dct)
    return html


def epmuchhtmllines(epmunch):
    """return the rows in the html table made from epmunch"""
    lines = 0
    for field in epmunch.epjfieldnames():
        if type(epmunch[field]) is list:
            lines += len(epmunch[field])
        else:
            lines += 1
    return lines


def epjhtmllines(epj):
    """return the number of rows in all the epmunch tables
    that make up the epj"""
    lines = 0
    for key in epj.epj.keys():
        for epmunch in epj.epobjects[key]:
            lines += epmuchhtmllines(epmunch)
    return lines


def epobjectslines(epobjects):
    """return the number of rows in all the epmunch tables
    that make up the epobjects"""
    lines = 0
    for epmunch in epobjects:
        lines += epmuchhtmllines(epmunch)
    return lines


def epmunch2ipythonhtml(epmunch, fname="./eppy3000_deletethis.html"):
    """display the epmunch as HTML table in jupyter notebook"""
    htmllines = epmuchhtmllines(epmunch)
    html = epmunch2html(epmunch)
    open(fname, "wb").write(html.encode())
    try:
        return IFrame(src=fname, width=800, height=30 * htmllines + 50)
    except NameError as e:
        raise JupyterNotInstalled


def epj2html(epj):
    """convert the epj to html"""
    epjjsonstr = epj.jsonstr()
    html = json2html.json2html.convert(json=epjjsonstr)
    return html


def epj2ipythonhtml(epj, fname="./eppy3000_deletethis.html"):
    """display the epj as HTML tables in jupyter notebook"""
    html = epj2html(epj)
    open(fname, "wb").write(html.encode())
    lines = epjhtmllines(epj)
    height = 30 * lines + 50
    try:
        return IFrame(src=fname, width=8000, height=height)
    except NameError as e:
        raise JupyterNotInstalled


def epobjects2dct(epobjects):
    """convert epobjects to dcts for display"""
    dct = {}
    for i, epmunch in enumerate(epobjects):
        eppykey = epmunch.eppykey
        eppyname = epmunch.eppyname
        displaymunch = _epmunch2displaymunch(epmunch)
        displaydict = displaymunch.toDict()
        for k1, val1 in displaydict.items():
            pass
        dct.setdefault(eppykey, dict())
        dct[eppykey].update(val1)
    return dct


def epobjects2html(epobjects):
    """convert epobjects to html"""
    dct = epobjects2dct(epobjects)
    html = json2html.json2html.convert(json=dct)
    return html


def epobjects2ipythonhtml(epobjects, fname="./eppy3000_deletethis.html"):
    """display the epobjects as HTML tables in jupyter notebook"""
    html = epobjects2html(epobjects)
    open(fname, "w").write(html)
    lines = epobjectslines(epobjects)
    height = 30 * lines + 50
    try:
        return IFrame(src=fname, width=800, height=height)
    except NameError as e:
        raise JupyterNotInstalled



# Copyright (c) 2019-2020 Santosh Philip
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
"""py.test for modelmaker"""


from io import StringIO
import json
import pytest

from eppy3000 import modelmaker
from tests import schemafortesting


@pytest.fixture
def version_txt():
    """txt of version"""
    txt = """{
    "Version": {
        "Version 1": {
            "version_identifier": "9.3",
            "idf_order": 1
        }
    },
    "SimulationControl": {},
    "Building": {},
    "Site:Location": {}
}

    """
    return txt


@pytest.fixture
def epj_txt1():
    """fixture for first epj"""
    txt = """
    {
        "BuildingSurface:Detailed": {
            "Zn001:Flr001": {
                "construction_name": "FLOOR",
                "idf_max_extensible_fields": 12,
                "idf_max_fields": 22,
                "idf_order": 27,
                "number_of_vertices": 4,
                "outside_boundary_condition": "Surface",
                "outside_boundary_condition_object": "Zn001:Flr001",
                "sun_exposure": "NoSun",
                "surface_type": "Floor",
                "vertices": [
                    {
                        "vertex_x_coordinate": 15.24,
                        "vertex_y_coordinate": 0.0,
                        "vertex_z_coordinate": 0.0
                    },
                    {
                        "vertex_x_coordinate": 0.0,
                        "vertex_y_coordinate": 0.0,
                        "vertex_z_coordinate": 0.0
                    },
                    {
                        "vertex_x_coordinate": 0.0,
                        "vertex_y_coordinate": 15.24,
                        "vertex_z_coordinate": 0.0
                    },
                    {
                        "vertex_x_coordinate": 15.24,
                        "vertex_y_coordinate": 15.24,
                        "vertex_z_coordinate": 0.0
                    }
                ],
                "view_factor_to_ground": 1.0,
                "wind_exposure": "NoWind",
                "zone_name": "Main Zone"
            }
        }
    }"""
    return txt


def test_EPJ(epj_txt1):
    """py.test for EPJ"""
    # test without epschema
    txt = epj_txt1
    expected = 15.24
    result = modelmaker.EPJ(epjname=StringIO(txt))
    surfs = result.epobjects["BuildingSurface:Detailed"]
    surf = surfs[0]
    assert expected == surf.vertices[0].vertex_x_coordinate

    # test with epschema
    expected = 15.24
    result = modelmaker.EPJ(
        epjname=StringIO(txt), epschemaname=schemafortesting.schema_file
    )
    surfs = result.epobjects["BuildingSurface:Detailed"]
    surf = surfs[0]
    assert expected == surf.vertices[0].vertex_x_coordinate

    # test jsonstr()
    expected = json.loads(txt)
    epj = modelmaker.EPJ(epjname=StringIO(txt))
    jsonstr = epj.jsonstr()
    result = json.loads(jsonstr)
    assert result == expected


def test_EPJ_repr_(version_txt):
    """pytest for __repr__ in EPJ"""
    expected = """
Version                                          !-  EP_KEY         # use .eppykey
            Version 1                            !-  EPJOBJECT_NAME # use .eppyname
            9.3                                  !-  version_identifier
            1                                    !-  idf_order"""
    epj = modelmaker.EPJ(epjname=StringIO(version_txt))
    assert expected == epj.__repr__()



# Copyright (c) 2018-2019 Santosh Philip
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
"""class for the epschema file"""

import json
from munch import Munch


def read_epschema_asmunch(fhandle):
    """read the epschema json as a munch"""
    try:
        epjs = json.load(fhandle)
        as_munch = EPSchemaMunch.fromDict(epjs)
    except AttributeError as e:
        try:
            fhandle = open(fhandle, "r")
            epjs = json.load(fhandle)
            as_munch = EPSchemaMunch.fromDict(epjs)
        except TypeError as e:
            if isinstance(fhandle, EPSchemaMunch):
                return fhandle
            else:
                raise TypeError(
                    f"expected str, bytes, os.PathLike object or Munch, not {type(fhandle)}"
                )  # noqa: E501
    return as_munch


class EPSchemaMunch(Munch):
    """Munch subcalssed to for the EPSchema json"""

    def __init__(self, *args, **kwargs):
        super(EPSchemaMunch, self).__init__(*args, **kwargs)

    def fieldnames(self):
        """field names of the EPSchema object"""
        return [key for key in self.keys()]

    def fieldnames_list(self):
        """fieldnames that contain lists"""
        pass

    def fieldproperty(self, fieldname):
        """field names of the EPSchema object"""
        return self[fieldname]


class EPSchema(object):
    """hold the data from the json epschema file """

    def __init__(self, epschemaname):
        super(EPSchema, self).__init__()
        self.epschemaname = epschemaname
        self.read()

    def read(self):
        """read the json file"""

        def prop_in_patternProp(val):
            """return the property in patternProperty"""
            # assume that val['patternProperties'] has a single key, val
            # key is either ".*" or "^.*\\S.*$"
            for key in val["patternProperties"].keys():
                return val["patternProperties"][key]

        self.epschema = read_epschema_asmunch(self.epschemaname)
        self.version = self.epschema["epJSON_schema_version"]
        self.required = self.epschema["required"]
        # self.epschemaobjects = {key: val['patternProperties']['.*']['properties']
        #                    for key, val in self.epschema['properties'].items()}
        # the above line stopped working in E+ V9.3. ".*" stopped being the only key
        # workaround is below
        self.epschemaobjects = {
            key: prop_in_patternProp(val)["properties"]
            for key, val in self.epschema["properties"].items()
        }



# Copyright (c) 2019 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================

"""py.test for rawidf"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from eppy3000 import rawidf
from io import StringIO


def test_readrawidf():
    """py.test for readrawidf"""
    # txt = """  version,9.0;
    # """
    txt = """  version,9.0;


    Timestep,4;

    ScheduleTypeLimits,
    Fraction,                !- Name
    0.0,                     !- Lower Limit Value
    1.0,                     !- Upper Limit Value
    CONTINUOUS;              !- Numeric Type

    ScheduleTypeLimits,
    Other,                !- Name
    0.0,                     !- Lower Limit Value
    1.0,                     !- Upper Limit Value
    CONTINUOUS;              !- Numeric Type
    """
    expected = {
        "version": [["version", "9.0"]],
        "Timestep": [["Timestep", "4"]],
        "ScheduleTypeLimits": [
            ["ScheduleTypeLimits", "Fraction", "0.0", "1.0", "CONTINUOUS"],
            ["ScheduleTypeLimits", "Other", "0.0", "1.0", "CONTINUOUS"],
        ],
    }
    result = rawidf.readrawidf(StringIO(txt))
    assert result == expected


def test_rawidf2str():
    """py.test rawidf2str"""
    rawdata = {"version".upper(): [["version", "9.0"]]}
    expected = """VERSION,
    9.0;
"""
    result = rawidf.rawidf2str(rawdata)
    assert result == expected
    rawdata = {
        "version".upper(): [["version", "9.0"]],
        "Timestep".upper(): [["Timestep", "4"]],
        "ScheduleTypeLimits".upper(): [
            ["ScheduleTypeLimits", "Fraction", "0.0", "1.0", "CONTINUOUS"],
            ["ScheduleTypeLimits", "Other", "0.0", "1.0", "CONTINUOUS"],
        ],
    }
    order = ["TIMESTEP", "VERSION", "SCHEDULETYPELIMITS", "badkey"]
    expected = """TIMESTEP,
    4;

VERSION,
    9.0;

SCHEDULETYPELIMITS,
     Fraction,
     0.0,
     1.0,
    CONTINUOUS;

SCHEDULETYPELIMITS,
     Other,
     0.0,
     1.0,
    CONTINUOUS;
"""
    result = rawidf.rawidf2str(rawdata, order)
    assert result == expected



# Copyright (c) 2019-2021 Santosh Philip
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
"""convertion functions - to convert from JSON to IDF and in reverse."""

import json
from itertools import zip_longest
import pathlib

from eppy3000 import rawidf
from eppy3000.epschema import read_epschema_asmunch
from eppy3000 import installlocation
from eppy3000.modelmaker import EPJ


def num(s):
    try:
        return int(s)
    except (ValueError, TypeError) as e:
        try:
            return float(s)
        except ValueError as e:
            return s


def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)


def keymapping(somekeys, allkeys):
    """map the keys regardless of case"""
    allkeys = list(allkeys)
    somekeys = list(somekeys)
    ll1 = [i.upper() for i in allkeys]
    ll2 = [i.upper() for i in somekeys]
    mapping = [(i, ll1.index(val)) for i, val in enumerate(ll2)]
    return {somekeys[i]: allkeys[j] for i, j in mapping}


def idf2json(idfhandle, epschemahandle):
    """converts the E+ file in the old IDF format to the new JSON format

    Parameters
    ----------
    jsonhandle: io.TextIOWrapper, io.StringIO
        This is the E+ file in the old IDF format
    epschemahandle: io.TextIOWrapper, io.StringIO
        This is the epjson file (eqv. of the IDD file in the old format)

    Returns
    -------
    str
        E+ file in the new JSON format
    """
    raw_idf = rawidf.readrawidf(idfhandle)
    js = read_epschema_asmunch(epschemahandle)
    idfobjcount = {}
    idfjson = {}
    keys = raw_idf.keys()
    order = 0
    mapping = keymapping(raw_idf.keys(), js.properties.keys())
    # mapping in case the case does not match between keys
    for akey in keys:
        key = mapping[akey]
        idfobjcount.setdefault(key, 0)
        dct = idfjson.setdefault(key, dict())
        fieldnames = js.properties[key].legacy_idd.fields
        idfobjects = raw_idf[akey]
        for idfobject in idfobjects:
            idfobjcount[key] = idfobjcount[key] + 1
            order += 1
            try:
                if fieldnames[0] == "name":
                    alst = {
                        fieldname: idfvalue
                        for idfvalue, fieldname in zip(idfobject[2:], fieldnames[1:])
                    }
                    idfobjectname = idfobject[1]
                else:
                    alst = {
                        fieldname: idfvalue
                        for idfvalue, fieldname in zip(idfobject[1:], fieldnames)
                    }
                    idfobjectname = f"{key} {idfobjcount[key]}"
            except IndexError as e:
                # catches "if fieldnames[0] == 'name':" when fieldnames = []
                alst = {
                    fieldname: idfvalue
                    for idfvalue, fieldname in zip(idfobject[1:], fieldnames)
                }
                idfobjectname = f"{key} {idfobjcount[key]}"
            alst["idf_order"] = order
            numericfields = js.properties[key].legacy_idd.numerics.fields
            for fieldkey in alst.keys():
                if fieldkey in numericfields:
                    alst[fieldkey] = num(alst[fieldkey])

            try:
                extension = js.properties[key].legacy_idd.extension
                extensibles = js.properties[key].legacy_idd.extensibles
                endvalues = idfobject[len(fieldnames) + 1 :]
                g_endvalues = grouper(endvalues, len(extensibles))
                extvalues = [
                    {f: t for f, t in zip(extensibles, tup)} for tup in g_endvalues
                ]

                try:
                    legacyidd = js.properties[key].legacy_idd
                    e_numericfields = legacyidd.numerics.extensions
                    for e_dct in extvalues:
                        for e_key in e_dct:
                            if e_key in e_numericfields:
                                if e_dct[
                                    e_key
                                ]:  # will skip None and '' ->maybe make sure of this
                                    e_dct[e_key] = num(e_dct[e_key])
                except AttributeError as e:
                    pass

                alst[extension] = extvalues
            except AttributeError as e:
                pass
            dct.update({idfobjectname: alst})
    return json.dumps(idfjson, indent=2)


def json2idf(jsonhandle, epschemahandle):
    """converts the E+ file new JSON format to the old IDF format

    Parameters
    ----------
    jsonhandle: io.TextIOWrapper, io.StringIO
        This is the E+ file in the new JSON format
    epschemahandle: io.TextIOWrapper, io.StringIO
        This is the epjson file (eqv. of the IDD file in the old format)

    Returns
    -------
    str
        E+ file in the old IDF format
    """
    lines = []
    js = read_epschema_asmunch(epschemahandle)
    idfjs = read_epschema_asmunch(jsonhandle)

    for key in idfjs.keys():
        for name in idfjs[key].keys():
            fieldval = []
            fieldnames = js.properties[key].legacy_idd.fields
            lastfield = len(fieldnames) - 1
            comma = ","
            semicolon = ";"
            sep = comma
            for i, fieldname in enumerate(fieldnames):
                if i == lastfield:
                    sep = semicolon
                try:
                    value = idfjs[key][name][fieldname]
                    fieldval.append((fieldname, value))
                except KeyError as e:
                    if fieldname == "name":
                        fieldval.append((fieldname, name))
                    else:
                        value = None
                        # debugging here
                        # fieldval.append((fieldname, value))
                        fieldval.append((fieldname, ""))
            try:
                extension = js.properties[key].legacy_idd.extension
                extensibles = js.properties[key].legacy_idd.extensibles
                for i, tup in enumerate(idfjs[key][name][extension]):
                    for fld in extensibles:
                        try:
                            fieldval.append((f"{fld} {i + 1}", tup[fld]))
                        except KeyError as e:
                            fieldval.append((f"{fld} {i + 1}", ""))
                        # fieldval.append((f"{fld} {i + 1}", tup[fld]))
            except AttributeError as e:
                pass
            fieldval = [(fld, val) for fld, val in fieldval if val is not None]
            # remove trailing blanks
            fieldval = removetrailingblanks(fieldval)
            # fieldval.reverse()
            # fieldval = [for first, second, in fieldval, if second]

            lastfield = len(fieldval) - 1
            sep = comma
            lines.append(f"{key},")
            for i, (fld, val) in enumerate(fieldval):
                if i == lastfield:
                    sep = semicolon
                valsep = f"{val}{sep}"
                lines.append(f"    {valsep:<25} !- {fld}")
            lines.append("")

    return "\n".join(lines)


def getidfversion(fhandle):
    """get the idf version of this file
    
    reads the Version object in the file and extract the version number

    Parameters
    ----------
    fhandle: io.String, io.TextIOWrapper
        can be a file open for read or a io.StringIO object

    Returns
    -------
    str
        Version number of the idf file"""
    foundword_version = False
    lines = []
    for line in fhandle:
        cline = line.split("!")[0]
        if foundword_version:
            if ";" in cline:
                remains = cline.split(";")[0]
                lines.append(remains)
                break
            else:
                lines.append(cline.strip())
        else:
            if "VERSION" in cline.upper():
                foundword_version = True
                fromversion = cline[cline.upper().find("VERSION") :].strip()
                if ";" in fromversion:
                    vline = fromversion.split(";")[0]
                    version = vline.split(",")[1].strip()
                    return version
                else:
                    lines.append(fromversion)
    compose = "".join(lines)
    if compose:
        return compose.split(",")[1].strip()
    else:
        return None


def idffile2epjfile(idfpath, epjpath=None, schemapath=None, epjext=None):
    """convert an IDF file on disk to an EPJ file on disk
    
    reads the idf file at idfpath. Converts it to epj file
    Writes the epj file to epjpath
    
    Parameters
    ----------
    idfpath: str, pathlib.Path
        idf file name or path to the idf file
    epjpath: str, pathlib.Path
        epj file name or path to the epj file
    schemapath: str, pathlib.Path
        path to Energy+.schema.epJSON usuallty found in the install location of EnergyPlus. If schemapah=None, the function will attempt to find the schema path in the standard install locations

    Returns
    -------
    pathlib.Path
        the path to the saved EPJ file
    """
    if not epjext:
        epjext = "epJSON"
    idfpath = pathlib.Path(idfpath)
    if not schemapath:
        with open(idfpath, "r") as idfhandle:
            version = getidfversion(idfhandle)
        schemapath = installlocation.schemapath(version)
    if not epjpath:
        epjpath = idfpath.with_suffix(f".{epjext}")
    schemahandle = open(schemapath, "r")
    idfhandle = open(idfpath, "r")
    epjtxt = idf2json(idfhandle, schemahandle)
    with open(epjpath, "w") as epjhandle:
        epjhandle.write(epjtxt)
    return epjpath


def idffolder2epjfolder(
    idffolder, epjfolder=None, idfext=None, epjext=None, schemapath=None, verbose=False
):
    """Convert all the idf files in idffolder into epj files and save them in epjfolder

    The function finds the idf files by searching for the all files in the idffolder that have the extension 'idf'. This extension can be overridden by giving a value to idfext

    Parameters
    ----------
    idffolder: str, pathlib.Path
        This is the folder with the idf files
    epjfolder: str, pathlib.Path
        This is the folder where you want to store the generated epj files. When epjfolder=None, the epj files are saved in the idffolder
    idfext: str
        This is the extension of the idf files. Default value is `idf`
    epjext: str
        This is the extension of the epj files. Default value is `epJSON`
    schemapath: str, pathlib.Path
        path to Energy+.schema.epJSON usuallty found in the install location of EnergyPlus. If schemapah=None, the function will attempt to find the schema path in the standard install locations
    verbose: Boolean
        When verbose=True, it will print file names on screen. The default for verbose is False

    Returns
    -------
    None
    
"""
    idffolder = pathlib.Path(idffolder)
    if not idfext:
        idfext = "idf"
    if not epjext:
        epjext = "epJSON"
    if not epjfolder:
        epjfolder = idffolder
    else:
        epjfolder = pathlib.Path(epjfolder)
    idfpaths = getfilepaths(idffolder, idfext)
    for idfpath in idfpaths:
        idfname = idfpath.name
        epjpath = epjfolder / idfname
        epjpath = epjpath.with_suffix(f".{epjext}")
        if verbose:
            print(f"converted {idfpath} -> {epjpath}")
        idffile2epjfile(idfpath, epjpath, schemapath)


def epjfile2idffile(epjpath, idfpath=None, schemapath=None):
    """convert an EPJ file on disk to an IDF file on disk"""
    epjpath = pathlib.Path(epjpath)
    if not schemapath:
        epj = EPJ(epjpath)
        version = epj.epobjects["Version"][0].version_identifier
        schemapath = installlocation.schemapath(version)
    if not idfpath:
        idfpath = epjpath.with_suffix(".idf")
    schemahandle = open(schemapath, "r")
    epjhandle = open(epjpath, "r")
    idftxt = json2idf(epjhandle, schemahandle)
    with open(idfpath, "w") as idfhandle:
        idfhandle.write(idftxt)


def removetrailingblanks(lst):
    """remove railing blanks in lst
    lst = [(a, b), (a, ''), (b, c), (d, ''), (e, '')]
    return [(a, b), (a, ''), (b, c)]"""
    lst.reverse()
    new_lst = []
    trailing = True
    for fst, snd in lst:
        if trailing:
            if snd != "":
                new_lst.append((fst, snd))
                trailing = False
            else:
                pass
        else:
            new_lst.append((fst, snd))
    new_lst.reverse()
    return new_lst


def getfilepaths(folder, ext):
    """get file paths of files in folder that have extension ext"""
    folder = pathlib.Path(folder)
    paths = list(folder.rglob(f"*.{ext}"))
    return [p for p in paths if p.parent == folder]


# def f_getidfversion(fhandle):
#     """functional version of getidfversion - an attempt"""
#
#     def aux(lines):
#         for line in lines:
#             cline = line.split('!')[0]
#             if not acc:
#                 if "VERSION" in cline.upper():
#                     if ";" in cline:
#                         version = cline.split(",")[1].strip()
#                         return None
#                     else:
#                         acc.append(cline.strip())
#                 else: pass
#             else:
#                 if ";" in cline:
#                     remains = cline.split(";")[0]
#                     acc.append(remains)
#                     break
#
#     acc = []
#     return aux(fhandle)
#     # what = sum([aux(line.split('!')[0], acc) for line in fhandle if ])
#     return what

# links to figure this out. Look into
#
# itertools.takewhile()
# https://stackoverflow.com/questions/9572833/using-break-in-a-list-comprehension
#
# from itertools import ifilter
# https://www.daniweb.com/programming/software-development/threads/293381/break-a-list-comprehension
#
# from itertools import takewhile
# from functools import partial
# https://www.reddit.com/r/learnpython/comments/7r3q1a/break_out_of_list_comprehension/
#



# Copyright (c) 2020 Santosh Philip
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
"""functions to work with list fields"""


def surf2list(surfobject):
    """return a list of xyz coords from the epobject"""
    vertices = surfobject["vertices"]
    return [
        (
            vrt["vertex_x_coordinate"],
            vrt["vertex_y_coordinate"],
            vrt["vertex_z_coordinate"],
        )
        for vrt in vertices
    ]



# Copyright (c) 2020 Santosh Philip
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
"""convertion functions - to convert from JSON to IDF and in reverse.
These functions return a eppy.IDF or eppy3000.EPJ structure"""


class NeedsEppyError(Exception):
    pass


try:
    import eppy
except ModuleNotFoundError as e:
    raise NeedsEppyError("you need to install eppy to run these functions")


from io import StringIO
import eppy3000
from eppy3000.modelmaker import EPJ
import eppy3000.idfjsonconverter

# -
from eppy import modeleditor
from eppy.modeleditor import IDF


def idf2epj(idf, epjsonhandle):
    """convert idf to json. return a eppy3000.IDF structure"""
    idfhandle = StringIO(idf.idfstr())
    epjstr = eppy3000.idfjsonconverter.idf2json(idfhandle, epjsonhandle)
    epjsonhandle.seek(0)  # need to reset, since we are reading it again
    return EPJ(StringIO(epjstr), epw=idf.epw, epschemaname=epjsonhandle)


def epj2idf(epj, epjsonhandle, iddhandle=None):
    """convert json to idf"""
    jsonhandle = epj.savecopy()
    idfstr = eppy3000.idfjsonconverter.json2idf(jsonhandle, epjsonhandle)
    return eppy.openidf(StringIO(idfstr), idd=iddhandle, epw=epj.epw)



# Copyright (c) 2021 Santosh Philip
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================

"""functions to find the installation folder of a Energyplus version"""

import platform


def version2folder(version):
    """foldername from version"""
    verlst = version.split(".")
    while len(verlst) < 3:
        verlst.append("0")
    return "-".join(verlst)


def installfolder(version, platform_system=None):
    """return the install folder of this EnergyPlus version"""
    if not platform_system:
        platform_system = platform.system()
    if platform_system == "Windows":
        pathstart = "C:/EnergyPlusV"
    elif platform_system == "Linux":
        pathstart = "/usr/local/EnergyPlus-"
    else:
        pathstart = "/Applications/EnergyPlus-"
    return f"{pathstart}{version2folder(version)}"


def schemapath(version, platform_system=None):
    """return the schema path for the Energyplus version"""
    installfolderpath = installfolder(version, platform_system)
    return f"{installfolderpath}/Energy+.schema.epJSON"



"""py.test for listfields"""

import pytest
from io import StringIO

from eppy3000.experimental import listfields
from eppy3000.modelmaker import EPJ


@pytest.mark.parametrize(
    "surfobjecttxt, expected",
    [
        (
            """{ "BuildingSurface:Detailed": {
        "BACK-1": {
            "construction_name": "WALL-1",
            "idf_max_extensible_fields": 12,
            "idf_max_fields": 22,
            "idf_order": 101,
            "number_of_vertices": 4,
            "outside_boundary_condition": "Outdoors",
            "sun_exposure": "SunExposed",
            "surface_type": "Wall",
            "vertices": [
                {
                    "vertex_x_coordinate": 30.5,
                    "vertex_y_coordinate": 15.2,
                    "vertex_z_coordinate": 2.4
                },
                {
                    "vertex_x_coordinate": 1,
                    "vertex_y_coordinate": 2,
                    "vertex_z_coordinate": 3
                }
            ],
            "view_factor_to_ground": 0.5,
            "wind_exposure": "WindExposed",
            "zone_name": "SPACE3-1"
        }
}
}""",
            [(30.5, 15.2, 2.4), (1, 2, 3)],
        ),  # surfobjecttxt, expected
    ],
)
def test_surf2list(surfobjecttxt, expected):
    """py.test for surf2list"""
    epj = EPJ(StringIO(surfobjecttxt))
    surfs = epj.epobjects["BuildingSurface:Detailed"]
    surfobject = surfs[0]
    result = listfields.surf2list(surfobject)
    assert result == expected



"""these test have to be run manually since eppy may not be installed"""

from io import StringIO
from pathlib import Path
import os

# from eppy3000 import idfjsonconverter
from tests import schemafortesting
from eppy3000.modelmaker import EPJ

# import eppy3000.oldeppy.idfjsonconverter as idfjsonconverter
import eppy3000.oldeppy as oldeppy
import eppy

SCHEMA_FILE = schemafortesting.schema_file

# path for iddfile -> clean this later
THIS_DIR = Path(os.path.dirname(os.path.abspath(__file__)))
IDDFILE = THIS_DIR / os.pardir / "eppy3000/resources/snippets/V9_1/iddV9_1_snippet.idd"
iddfile = IDDFILE


def test_idf2idj_epj2idf():
    """py.test for idf2idj and epj2idf"""
    # rearranged the order of so that it matches the order in the idd file.
    # So the tests can pass
    data = (
        (
            """
  Version,9.1;

Building,
    Bldg one,                !- Name
    30,                     !- North Axis {deg}
    City,                    !- Terrain
    0.04,                    !- Loads Convergence Tolerance Value
    0.4,                     !- Temperature Convergence Tolerance Value {deltaC}
    FullExterior,            !- Solar Distribution
    25,                      !- Maximum Number of Warmup Days
    6;                       !- Minimum Number of Warmup Days

Timestep,4;

Zone,
    Main Zone,               !- Name
    0,                       !- Direction of Relative North {deg}
    0,                       !- X Origin {m}
    0,                       !- Y Origin {m}
    0,                       !- Z Origin {m}
    1,                       !- Type
    1,                       !- Multiplier
    autocalculate,           !- Ceiling Height {m}
    autocalculate;           !- Volume {m3}

Zone,
    Another Zone,               !- Name
    90,                       !- Direction of Relative North {deg}
    11,                       !- X Origin {m}
    22,                       !- Y Origin {m}
    33,                       !- Z Origin {m}
    1,                       !- Type
    1,                       !- Multiplier
    autocalculate,           !- Ceiling Height {m}
    autocalculate;           !- Volume {m3}


BuildingSurface:Detailed,
    WALL-1PF,                !- Name
    WALL,                    !- Surface Type
    WALL-1,                  !- Construction Name
    PLENUM-1,                !- Zone Name
    Outdoors,                !- Outside Boundary Condition
    ,                        !- Outside Boundary Condition Object
    SunExposed,              !- Sun Exposure
    WindExposed,             !- Wind Exposure
    0.50000,                 !- View Factor to Ground
    4,                       !- Number of Vertices
    0,0,3,  !- X,Y,Z ==> Vertex 1 {m}
    0,0,2.4,  !- X,Y,Z ==> Vertex 2 {m}
    30.5,0,2.4,  !- X,Y,Z ==> Vertex 3 {m}
    30.5,0,3;  !- X,Y,Z ==> Vertex 4 {m}


  ZoneHVAC:EquipmentList,
    SPACE1-1 Eq,             !- Name
    SequentialLoad,          !- Load Distribution Scheme
    ZoneHVAC:AirDistributionUnit,  !- Zone Equipment 1 Object Type
    SPACE1-1 ATU,            !- Zone Equipment 1 Name
    1,                       !- Zone Equipment 1 Cooling Sequence
    1;                       !- Zone Equipment 1 Heating or No-Load Sequence


NodeList,
    SPACE1-1 In Nodes,       !- Name
    SPACE1-1 In Node;        !- Node 1 Name

NodeList,
    Supply Air Temp Nodes 1, !- Name
    Main Cooling Coil 1 Inlet Node,  !- Node 1 Name
    Heat Recovery Supply Outlet,  !- Node 2 Name
    Main Heating Coil 1 Inlet Node,  !- Node 3 Name
    Reheat Coil Air Inlet Node,  !- Node 4 Name
    VAV Sys 1 Outlet Node;   !- Node 5 Name

OutdoorAir:NodeList,
    OutsideAirInletNodes;    !- Node or NodeList Name 1

""",  # noqa: E501
        ),  # idftxt
    )
    for (idftxt,) in data:
        iddhandle = open(iddfile, "r")
        iddtxt = iddhandle.read()
        iddstringio = StringIO(iddtxt)

        # convert idf to epj, then epj to idf and lastly idf to epj
        # compare the first epj.json and last epj.json
        idfhandle = StringIO(idftxt)
        idf = eppy.openidf(idfhandle, idd=iddstringio)
        epsjsonschema = schemafortesting.schema

        # idf to epj
        epj1 = oldeppy.idf2epj(idf, epsjsonschema)

        # epj to idf
        idf1 = oldeppy.epj2idf(epj1, epsjsonschema, iddhandle=iddstringio)

        # idf to epj
        epj2 = oldeppy.idf2epj(idf1, epsjsonschema)

        # compare epj1 and epj2
        assert epj1.jsonstr() == epj2.jsonstr()



# Copyright (c) 2018-2019 Santosh Philip
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
"""subclass Munch"""

from munch import Munch
from eppy3000.epschema import EPSchemaMunch


class NotEPObject(Exception):
    pass


def printkey(key, indent=0, formatstr=None, func=None):
    """Prints the key in epMunch with the right indentation

    Used internally by printmunch"""
    if not func:
        func = print
    if not indent:
        func("")
        func("")
        indent = 0
    elif indent == 1:
        func("")
    if not formatstr:
        formatstr = "{}{}"
        func(formatstr.format(" " * 4 * indent, key))
    else:
        func(formatstr.format(" " * 4 * indent, key))


def printmunch(amunch, indent=0, index=None, func=None):
    """This prints the epMunch object

    In effect, it can print the epJSON file
    which may have nested epMunch objects in it. Or it can
    print a single epMunch object.
    printmunch is called recursively until all epMunch
    objects are exhausted. It will also  print a list of
    epMunch objects that can occur within an epJSON file

    It has been tested for epJSON files. The is no guarantee
    that will work on a more complex nesting of epMunch objects

    An epJSON object such as::


         "Heating Setpoint Schedule": {
             "data": [
                 {
                     "field": "Through: 12/31"
                 },
                 {
                     "field": "For: AllDays"
                 },
                 {
                     "field": "Until: 24:00"
                 },
                 {
                     "field": 15.0
                 }
             ],
             "idf_max_extensible_fields": 4,
             "idf_max_fields": 6,
             "idf_order": 59,
             "schedule_type_limits_name": "Any Number"
         },

    will print out as::


        Schedule:Compact                                 !-  EP_KEY
                    Heating Setpoint Schedule            !-  EPJOBJECT_NAME
                                                         !-  data
                        Through: 12/31                       !-  field #1
                        For: AllDays                         !-  field #2
                        Until: 24:00                         !-  field #3
                        15.0                                 !-  field #4
                    4                                    !-  idf_max_extensible_fields
                    6                                    !-  idf_max_fields
                    59                                   !-  idf_order
                    Any Number                           !-  schedule_type_limits_name

    which is much more easy on human eyes

    Parameters
    ----------
    amunch: epMunch
        an EpMunch object that may have more nested EpMunch objects within
    indent: int
        Used internally to indent the output
    index: int
        used internally to give index numbers to repeating items in a list
    func: function
        default func = print
        you can get a list by::

            lines = []
            printmunch(amunch, func=line.append)
            # lines will be a list of lines

    Returns
    -------
    None
    """  # noqa: E501
    if isinstance(amunch, EPSchemaMunch):  # don't print EPSchema stuff
        return
    if not func:
        func = print
    if isinstance(amunch, Munch):
        if "eppykey" in amunch:
            func("")
            func(
                f"{amunch['eppykey']:<36}{' '*4*(indent+1)} !-  EP_KEY         # use .eppykey"
            )
            ind1 = " " * 4 * (indent + 1)
            ind2 = " " * 4 * (indent - 1)
            func(
                f"{ind1}{amunch['eppyname']:<32} {ind2}!-  EPJOBJECT_NAME # use .eppyname"
            )
    for key, val in amunch.items():
        if key in ["eppy_epj", "eppy_epobjects"]:
            continue  # prevents an infinite recurse
        if isinstance(val, Munch):
            printmunch(val, indent=indent + 1, index=index, func=func)
        elif isinstance(val, list):
            printkey(key, indent=3, formatstr="{0}" + " " * 36 + " !-  {1}", func=func)
            for i, aval in enumerate(val):
                printmunch(aval, indent=indent + 1, index=i + 1, func=func)
        elif key not in [
            "eppykey",
            "eppyname",
            "eppy_obj_schema",
            "eppy_epj",
            "eppy_epobjects",
        ]:
            if index:
                ind = " " * 4 * (indent + 1)
                func(f"{ind}{val:<36} !-  {key} #{index}")
            else:
                ind = " " * 4 * (indent + 1)
                func(f"{ind}{val:<36} !-  {key}")


class EPMunch(Munch):
    """Subclass of Munch for eppy3000"""

    def __init__(self, *args, **kwargs):
        super(EPMunch, self).__init__(*args, **kwargs)

    # TODO : __repr__ and __str__ ahould be different
    def __repr__(self):
        """print this as a snippet"""
        lines = []
        printmunch(self, func=lines.append)
        return "\n".join(lines)

    def __str__(self):
        """same as __repr__"""
        return self.__repr__()

    def __setitem__(self, key, value):
        """only for epobjects and for key starting with eppy"""
        # EPMunchobj['keyname'] = value will
        #   call __setitem__
        # EPMunchobj.keyname = value will
        #   call __setattr__ and then call
        #   call __setitem__
        if "eppykey" not in self:
            # it is not an epobject. None of this applies
            super(EPMunch, self).__setitem__(key, value)
        else:
            # it is an epobject
            if key in self:
                if key.startswith("eppy"):  # TODO: test if it is a string
                    if key == "eppyname":
                        if key in self.keys():
                            # if 'eppyname' is changed, the following actions happen in the parent dict
                            # the old key (value of 'eppyname') is popped and a new key is added
                            # with the same value
                            # epobjects_dict = self.eppy_epj[self.eppykey]
                            epobjects_dict = self.eppy_epobjects
                            epobject = epobjects_dict.pop(self.eppyname)
                            epobjects_dict[value] = self
                            self.pop(key)
                            super(EPMunch, self).__setitem__(key, value)
                        else:
                            super(EPMunch, self).__setitem__(key, value)
                    else:
                        pass  # user is not allowed to change eppy fields
                else:
                    super(EPMunch, self).__setitem__(key, value)
            else:
                super(EPMunch, self).__setitem__(key, value)

    def __setattr__(self, name, value):
        """deals with names starting with eppy. Name starts with eppy only in epobjects"""
        # EPMunchobj['keyname'] = value will
        #   call __setitem__
        # EPMunchobj.keyname = value will
        #   call __setattr__ and then call
        #   call __setitem__
        if "eppykey" not in self:
            # it is not an epobject
            super(EPMunch, self).__setattr__(name, value)  # Let Munch handle it
        else:
            # it is an epobject
            if name.startswith("eppy"):
                if name in self:
                    pass  # if the key exists, user cannot change it
                else:
                    super(EPMunch, self).__setattr__(name, value)  # Let Munch handle it
            else:
                super(EPMunch, self).__setattr__(name, value)  # Let Munch handle it
            if name == "eppyname":
                if name in self.keys():
                    # if 'eppyname' is changed, the following actions happen in the parent dict
                    # the old key (value of 'eppyname') is popped and a new key is added
                    # with the same value
                    epobjects_dict = self.eppy_epobjects
                    epobject = epobjects_dict.pop(self.eppyname)
                    epobjects_dict[value] = self
                    self.pop(name)
                    super(EPMunch, self).__setattr__(name, value)  # Let Munch handle it
                else:
                    super(EPMunch, self).__setattr__(name, value)  # Let Munch handle it

    def delete(self):
        """delete this by removing it from the parent dict"""
        try:
            epobjects = self.eppy_epobjects
        except AttributeError as e:
            raise NotEPObject
        return epobjects.pop(self.eppyname)

    def copy(self, newname):
        """make a copy of this anc add it to parent dict"""
        pass

    def allfieldnames(self):
        return list(self.keys())

    def epjfieldnames(self):
        fieldnames = [key for key in self.keys() if not key.startswith("eppy")]
        return fieldnames

    def schemafieldnames(self):
        return list(self.eppy_objepschema.keys())



# Copyright (c) 2019 Santosh Philip
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================

"""py.test for epschema.py"""

from io import StringIO
from eppy3000 import epschema
import pytest

from tests import schemafortesting


def test_read_epschema_asmunch():
    """py.test for read_epschema_asmunch"""
    SCHEMA_FILE = schemafortesting.schema_file
    schemahandle = open(SCHEMA_FILE, "r")
    result = epschema.read_epschema_asmunch(schemahandle)
    assert isinstance(result, epschema.EPSchemaMunch)

    result = epschema.read_epschema_asmunch(SCHEMA_FILE)
    assert isinstance(result, epschema.EPSchemaMunch)

    schemahandle = schemafortesting.schema
    result = epschema.read_epschema_asmunch(schemahandle)
    assert isinstance(result, epschema.EPSchemaMunch)

    schemahandle = list()
    with pytest.raises(TypeError):
        result = epschema.read_epschema_asmunch(schemahandle)


def test_EPSchema():
    """py.test for EPSchema"""
    txt = """{
    "epJSON_schema_version": "8.9.0",
    "$schema": "http://json-schema.org/draft-04/schema#",
    "required": [
        "Building",
        "GlobalGeometryRules"
    ],
    "properties": {
        "Building": {
            "name": {
                "default": "NONE",
                "retaincase": true,
                "type": "string"
            },
            "memo": "Describes parameters that are used during the simulation of the building. There are necessary correlations between the entries for this object and some entries in the Site:WeatherStation and Site:HeightVariation objects, specifically the Terrain field.",
            "patternProperties": {
                ".*": {
                    "type": "object",
                    "properties": {
                        "solar_distribution": {
                            "note": "MinimalShadowing | FullExterior | FullInteriorAndExterior | FullExteriorWithReflections | FullInteriorAndExteriorWithReflections",
                            "default": "FullExterior",
                            "enum": [
                                "",
                                "FullExterior",
                                "FullExteriorWithReflections",
                                "FullInteriorAndExterior",
                                "FullInteriorAndExteriorWithReflections",
                                "MinimalShadowing"
                            ],
                            "type": "string"
                        },
                        "terrain": {
                            "note": "Country=FlatOpenCountry | Suburbs=CountryTownsSuburbs | City=CityCenter | Ocean=body of water (5km) | Urban=Urban-Industrial-Forest",
                            "default": "Suburbs",
                            "enum": [
                                "",
                                "City",
                                "Country",
                                "Ocean",
                                "Suburbs",
                                "Urban"
                            ],
                            "type": "string"
                        },
                        "north_axis": {
                            "note": "degrees from true North",
                            "units": "deg",
                            "default": 0.0,
                            "type": "number"
                        },
                        "maximum_number_of_warmup_days": {
                            "note": "EnergyPlus will only use as many warmup days as needed to reach convergence tolerance. This field's value should NOT be set less than 25.",
                            "exclusiveMinimum": true,
                            "default": 25.0,
                            "minimum": 0.0,
                            "type": "number"
                        },
                        "loads_convergence_tolerance_value": {
                            "default": 0.04,
                            "type": "number",
                            "maximum": 0.5,
                            "note": "Loads Convergence Tolerance Value is a fraction of load",
                            "minimum": 0.0,
                            "exclusiveMinimum": true
                        },
                        "temperature_convergence_tolerance_value": {
                            "default": 0.4,
                            "type": "number",
                            "maximum": 0.5,
                            "minimum": 0.0,
                            "units": "deltaC",
                            "exclusiveMinimum": true
                        },
                        "minimum_number_of_warmup_days": {
                            "note": "The minimum number of warmup days that produce enough temperature and flux history to start EnergyPlus simulation for all reference buildings was suggested to be 6. When this field is greater than the maximum warmup days defined previous field the maximum number of warmup days will be reset to the minimum value entered here. Warmup days will be set to be the value you entered when it is less than the default 6.",
                            "exclusiveMinimum": true,
                            "default": 6.0,
                            "minimum": 0.0,
                            "type": "number"
                        }
                    }
                }
            },
            "legacy_idd": {
                "numerics": {
                    "fields": [
                        "north_axis",
                        "loads_convergence_tolerance_value",
                        "temperature_convergence_tolerance_value",
                        "maximum_number_of_warmup_days",
                        "minimum_number_of_warmup_days"
                    ]
                },
                "field_info": {
                    "name": {
                        "field_type": "a",
                        "field_name": "Name"
                    },
                    "solar_distribution": {
                        "field_type": "a",
                        "field_name": "Solar Distribution"
                    },
                    "terrain": {
                        "field_type": "a",
                        "field_name": "Terrain"
                    },
                    "north_axis": {
                        "field_type": "n",
                        "field_name": "North Axis"
                    },
                    "maximum_number_of_warmup_days": {
                        "field_type": "n",
                        "field_name": "Maximum Number of Warmup Days"
                    },
                    "loads_convergence_tolerance_value": {
                        "field_type": "n",
                        "field_name": "Loads Convergence Tolerance Value"
                    },
                    "temperature_convergence_tolerance_value": {
                        "field_type": "n",
                        "field_name": "Temperature Convergence Tolerance Value"
                    },
                    "minimum_number_of_warmup_days": {
                        "field_type": "n",
                        "field_name": "Minimum Number of Warmup Days"
                    }
                },
                "alphas": {
                    "fields": [
                        "name",
                        "terrain",
                        "solar_distribution"
                    ]
                },
                "fields": [
                    "name",
                    "north_axis",
                    "terrain",
                    "loads_convergence_tolerance_value",
                    "temperature_convergence_tolerance_value",
                    "solar_distribution",
                    "maximum_number_of_warmup_days",
                    "minimum_number_of_warmup_days"
                ]
            },
            "maxProperties": 1,
            "minProperties": 1,
            "type": "object",
            "min_fields": 8.0
        },
        "OutdoorAir:NodeList": {
            "type": "object",
            "memo": "This object sets the temperature and humidity conditions for an outdoor air node using the weather data values. to vary outdoor air node conditions with height above ground use OutdoorAir:Node instead of this object. This object may be used more than once. The same node name may not appear in both an OutdoorAir:Node object and an OutdoorAir:NodeList object.",
            "legacy_idd": {
                "numerics": {
                    "fields": []
                },
                "field_info": {
                    "node_or_nodelist_name": {
                        "field_type": "a",
                        "field_name": "Node or NodeList Name"
                    }
                },
                "extensibles": [
                    "node_or_nodelist_name"
                ],
                "extension": "nodes",
                "alphas": {
                    "fields": [],
                    "extensions": [
                        "node_or_nodelist_name"
                    ]
                },
                "fields": []
            },
            "patternProperties": {
                ".*": {
                    "type": "object",
                    "properties": {
                        "nodes": {
                            "items": {
                                "required": [
                                    "node_or_nodelist_name"
                                ],
                                "type": "object",
                                "properties": {
                                    "node_or_nodelist_name": {
                                        "type": "string"
                                    }
                                }
                            },
                            "type": "array"
                        }
                    }
                }
            },
            "extensible_size": 1.0
        }
    },
    "epJSON_schema_build": "40101eaafd"
}"""  # noqa: E501

    # expected
    version = "8.9.0"
    required = ["Building", "GlobalGeometryRules"]
    epschemaobjectskeys = ["Building", "OutdoorAir:NodeList"]
    buildingkeys = [
        "solar_distribution",
        "terrain",
        "north_axis",
        "maximum_number_of_warmup_days",
        "loads_convergence_tolerance_value",
        "temperature_convergence_tolerance_value",
        "minimum_number_of_warmup_days",
    ]
    buildingterraintype = "string"
    bfieldnames = [
        "solar_distribution",
        "terrain",
        "north_axis",
        "maximum_number_of_warmup_days",
        "loads_convergence_tolerance_value",
        "temperature_convergence_tolerance_value",
        "minimum_number_of_warmup_days",
    ]
    fieldnameslist = None  # not yet coded
    #
    # tests
    fhandle = StringIO(txt)
    result = epschema.EPSchema(fhandle)

    assert version == result.version
    assert required == result.required
    assert epschemaobjectskeys == list(result.epschemaobjects.keys())
    assert buildingkeys == list(result.epschemaobjects["Building"].keys())
    #
    bldg = result.epschemaobjects["Building"]
    assert buildingterraintype == bldg.fieldproperty("terrain")["type"]
    assert bfieldnames == result.epschemaobjects["Building"].fieldnames()
    oairnode = result.epschemaobjects["OutdoorAir:NodeList"]
    assert fieldnameslist == oairnode.fieldnames_list()



# Copyright (c) 2021 Santosh Philip
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
"""py.test for epjviewer"""

import pytest
from eppy3000.modelmaker import EPJ
from eppy3000 import epjviewer

try:
    from IPython.display import IFrame

    JUPYTER = True
except ModuleNotFoundError as e:
    JUPYTER = False

# schema_file = "./eppy3000/resources/schema/V9_3/Energy+.schema.epJSON"
ep_file = "./eppy3000/resources/epJSON/V9_3/smallfile.epJSON"
epj = EPJ(epjname=ep_file)
ep_file_big = (
    "./eppy3000/resources/snippets/V9_0/5Zone_Unitary_HXAssistedCoil.epJSONout"
)
epj_big = EPJ(epjname=ep_file_big)


def test_epj2html():
    """py.test for epj2html"""
    expected = '<table border="1"><tr><th>Version</th><td><table border="1"><tr><th>Version 1</th><td><table border="1"><tr><th>version_identifier</th><td>9.3</td></tr><tr><th>idf_order</th><td>1</td></tr></table></td></tr></table></td></tr><tr><th>SimulationControl</th><td><table border="1"><tr><th>SimulationControl 1</th><td><table border="1"><tr><th>do_zone_sizing_calculation</th><td>Yes</td></tr><tr><th>do_system_sizing_calculation</th><td>Yes</td></tr><tr><th>do_plant_sizing_calculation</th><td>Yes</td></tr><tr><th>run_simulation_for_sizing_periods</th><td>No</td></tr><tr><th>run_simulation_for_weather_file_run_periods</th><td>Yes</td></tr><tr><th>idf_order</th><td>2</td></tr></table></td></tr></table></td></tr><tr><th>Building</th><td><table border="1"><tr><th>Empire State Building</th><td><table border="1"><tr><th>north_axis</th><td>30</td></tr><tr><th>terrain</th><td>City</td></tr><tr><th>loads_convergence_tolerance_value</th><td>0.04</td></tr><tr><th>temperature_convergence_tolerance_value</th><td>0.4</td></tr><tr><th>solar_distribution</th><td>FullExterior</td></tr><tr><th>maximum_number_of_warmup_days</th><td>25</td></tr><tr><th>minimum_number_of_warmup_days</th><td>6</td></tr><tr><th>idf_order</th><td>3</td></tr></table></td></tr></table></td></tr><tr><th>Site:Location</th><td><table border="1"><tr><th>CHICAGO_IL_USA TMY2-94846</th><td><table border="1"><tr><th>latitude</th><td>41.78</td></tr><tr><th>longitude</th><td>-87.75</td></tr><tr><th>time_zone</th><td>-6</td></tr><tr><th>elevation</th><td>190</td></tr><tr><th>idf_order</th><td>4</td></tr></table></td></tr></table></td></tr></table>'
    result = epjviewer.epj2html(epj)
    assert result == expected


def test_epmunch2dct():
    """py.test for epmunch2dct"""
    expected = {"Version": {"Version 1": {"idf_order": 1, "version_identifier": "9.3"}}}
    versions = epj.epobjects["Version"]
    version = versions[0]
    result = epjviewer.epmunch2dct(version)
    assert result == expected


def test_epmunch2html():
    """py.test for epmunch2html"""
    expected = '<table border="1"><tr><th>Version</th><td><table border="1"><tr><th>Version 1</th><td><table border="1"><tr><th>version_identifier</th><td>9.3</td></tr><tr><th>idf_order</th><td>1</td></tr></table></td></tr></table></td></tr></table>'
    versions = epj.epobjects["Version"]
    version = versions[0]
    result = epjviewer.epmunch2html(version)
    assert result == expected


def test_epmunch2ipythonhtml():
    """py.test for epmunch2ipythonhtml"""
    versions = epj.epobjects["Version"]
    version = versions[0]
    if JUPYTER:
        expected = IFrame
        result = epjviewer.epmunch2ipythonhtml(version)
        assert type(result) == expected
    else:
        with pytest.raises(epjviewer.JupyterNotInstalled):
            result = epjviewer.epmunch2ipythonhtml(version)


def test_epmuchhtmllines():
    """py.test for epmuchhtmllines"""
    surfaces = epj_big.epobjects["BuildingSurface:Detailed"]
    surface = surfaces[0]
    result = epjviewer.epmuchhtmllines(surface)
    expected = 15
    assert result == expected


def test_epjhtmllines():
    """py.test for epjhtmllines"""
    result = epjviewer.epjhtmllines(epj)
    expected = 21
    assert result == expected


def test_epobjectslines():
    """py.test for epobjectslines"""
    surfs = epj_big.epobjects["BuildingSurface:Detailed"]
    roofs = [surf for surf in surfs if surf.surface_type == "Roof"]
    version = epj_big.epobjects["Version"]
    epobjects = version + roofs
    result = epjviewer.epobjectslines(epobjects)
    expected = 19
    assert result == expected


def test_epj2ipythonhtml():
    """py.test for epj2ipythonhtml"""
    if JUPYTER:
        result = epjviewer.epj2ipythonhtml(epj)
        expected = IFrame
        assert type(result) == expected  # just ensure that it runs
    else:
        with pytest.raises(epjviewer.JupyterNotInstalled):
            result = epjviewer.epj2ipythonhtml(epj)


def test_epobjects2dct():
    """py.test for epobjects2dct"""
    epobjects = epj_big.epobjects["Version"]
    result = epjviewer.epobjects2dct(epobjects)
    expected = {
        "Version": {
            "Version 1": {
                "idf_max_extensible_fields": 0,
                "idf_max_fields": 1,
                "idf_order": 1,
                "version_identifier": "9.0",
            }
        }
    }
    assert result == expected


def test_epobjects2html():
    """py.test for epobjects2html"""
    epobjects = epj_big.epobjects["Version"]
    result = epjviewer.epobjects2html(epobjects)
    expected = '<table border="1"><tr><th>Version</th><td><table border="1"><tr><th>Version 1</th><td><table border="1"><tr><th>idf_max_extensible_fields</th><td>0</td></tr><tr><th>idf_max_fields</th><td>1</td></tr><tr><th>idf_order</th><td>1</td></tr><tr><th>version_identifier</th><td>9.0</td></tr></table></td></tr></table></td></tr></table>'
    assert result == expected


def test_epobjects2ipythonhtml():
    """py.test for epobjects2ipythonhtml"""
    epobjects = epj_big.epobjects["Version"]
    if JUPYTER:
        expected = IFrame
        result = epjviewer.epobjects2ipythonhtml(epobjects)
        assert type(result) == expected  # just ensure that it runs
    else:
        with pytest.raises(epjviewer.JupyterNotInstalled):
            result = epjviewer.epobjects2ipythonhtml(epobjects)



# Copyright (c) 2018 Santosh Philip
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
"""same as modelmaker in eppy"""

from io import StringIO
from munch import Munch
from eppy3000.readepj import readepjjson
from eppy3000.epschema import read_epschema_asmunch
from eppy3000.readepj import removeeppykeys
from eppy3000.epschema import EPSchema
from eppy3000.epMunch import EPMunch
import eppy3000.runner.run_functions as run_functions


class EPJ(object):
    def __init__(self, epjname=None, epw=None, epschemaname=None):
        super(EPJ, self).__init__()
        self.epjname = epjname
        self.epw = epw
        self.epschemaname = epschemaname
        if self.epschemaname:
            self.readepschema()
        self.read()

    def read_epschema_asmunch(self):
        """Read the epschema file - will become a frozen singleton"""
        read_epschema_asmunch(self.epschemaname)
        # not used ???

    def read(self):
        """read the epj file"""
        self.epj = readepjjson(self.epjname)
        self.epobjects = {
            key: [val1 for val1 in val.values()] for key, val in self.epj.items()
        }
        # TODO: the above line should get the epobjects from the schema
        # This should happen whenever the schema is read.
        # in case the schema reading happens far in the future
        # eppy3000 should partilly work even without the schema
        # -

        # insert epj into the epjobject
        # this allows the epjobject to access all other objects in the epj
        for epobjects in self.epobjects.values():
            for epobject in epobjects:
                epobject["eppy_epj"] = self.epj
        if self.epschemaname:
            for key in self.epobjects.keys():
                for epobject in self.epobjects[key]:
                    epobject["eppy_objepschema"] = self.epschema.epschemaobjects[key]

    def readepschema(self):
        """read the epschema file"""
        self.epschema = EPSchema(self.epschemaname)

    def __repr__(self):
        """print this"""
        return self.epj.__repr__()

    def saveas(self, filename, indent=4):
        """saveas in filename"""
        self.epjname = filename
        self.save(filename, indent=indent)

    def savecopy(self, filename=None, indent=4):
        """save a copy of the file
        if filename==None: return copy in StringIO
        NOT unit TESTED at all. Not even use tested"""
        if filename:
            self.save(filename=filename, indent=indent)
        else:
            tosave = self.epj.toDict()
            tosave = Munch.fromDict(tosave)
            removeeppykeys(tosave)
            fhandle = StringIO()
            fhandle.write(tosave.toJSON(indent=indent))
            fhandle.seek(0)
            return fhandle

    def save(self, filename=None, indent=4):
        """save the file"""
        if not filename:
            filename = self.epjname
        with open(filename, "w") as fhandle:
            tosave = self.epj.toDict()
            tosave = Munch.fromDict(tosave)
            removeeppykeys(tosave)
            fhandle.write(tosave.toJSON(indent=indent))

    def jsonstr(self, indent=4):
        """return a json string of the file"""
        fhandle = self.savecopy(indent=indent)
        return "\n".join([line.rstrip() for line in fhandle])

    def newepobject(self, key, objname, defaultvalues=True, **kwargs):
        """create a new epj object"""
        # TODO test for dup name
        # TODO Kwargs strategy for array -
        #     delay implementation for now, throw exception
        # TODO exceptions for wrong field name
        # TODO exception for wrong field value type
        # TODO documentation in usage.rst
        # should self.epobjects be updated here
        objepschema = self.epschema.epschemaobjects[key]
        try:
            nobj = self.epj[key][objname] = EPMunch()
        except KeyError as e:
            self.epj[key] = EPMunch()
            nobj = self.epj[key][objname] = EPMunch()
        for fieldname in objepschema.fieldnames():
            try:
                if defaultvalues:
                    fieldfprop = objepschema.fieldproperty(fieldname)
                    nobj[fieldname] = fieldfprop["default"]
            except KeyError as e:
                prop = objepschema.fieldproperty(fieldname)
                # print(fieldname, prop.keys())
                if "type" in prop:
                    if prop["type"] == "array":
                        # pprint(prop['items'])
                        pass
                pass
        for key1, val1 in kwargs.items():
            nobj[key1] = val1
        nobj["eppykey"] = key
        nobj["eppyname"] = objname
        nobj["eppy_objepschema"] = objepschema
        return nobj

    def removeepobject(self, key, objname):
        """remove an epj object"""
        # should self.epobjects be updated here
        return self.epj[key].pop(objname)

    def copyepobject(self, key, objname, newname):
        """copy an epj object with a new name"""
        # don't use the function dict.items() since the json for array has
        # field name `items`
        # should self.epobjects be updated here
        oldobj = self.epj[key][objname]
        newobj = EPMunch()
        self.epj[key][newname] = newobj
        for key1 in oldobj.keys():
            if not key1.startswith("eppy"):
                val1 = oldobj[key1]
                if isinstance(val1, list):
                    newobj[key1] = list()
                    for item in val1:
                        newobj[key1].append(item.copy())
                else:
                    newobj[key1] = val1
        newobj["eppyname"] = newname
        newobj["eppykey"] = key
        newobj["eppy_objepschema"] = oldobj["eppy_objepschema"]
        return newobj

    def run(self, **runoptions):
        return run_functions.run(self, **runoptions)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()

requirements = ["munch", "eppy", "json2html"]

setup_requirements = [
    "pytest-runner",
]

test_requirements = [
    "pytest",
]

setup(
    author="Santosh Philip",
    author_email="santosh@noemail.com",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    description="E+ scripting using epJSON file format",
    install_requires=requirements,
    license="Mozilla Public License, v. 2.0",
    long_description=readme + "\n\n" + history,
    include_package_data=True,
    keywords="eppy3000",
    name="eppy3000",
    packages=find_packages(include=["eppy3000", "eppy3000.*"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/pyenergyplus/eppy3000",
    version="0.1.14",
    zip_safe=False,
)



# Copyright (c) 2019-2020 Santosh Philip
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
"""py.test for epMunch"""

import json
from io import StringIO
import pytest

import eppy3000.readepj as readepj
from eppy3000 import epMunch
from eppy3000.epschema import EPSchemaMunch
from munch import Munch
from eppy3000.readepj import removeeppykeys


def test_printkey():
    """py.test for printkey"""
    data = (
        ("Gumby", None, None, ["", "", "Gumby"]),  # key, indent, formatstr, expected
        ("Gumby", 1, None, ["", "    Gumby"]),  # key, indent, formatstr, expected
        ("Gumby", None, "{}{}", ["", "", "Gumby"]),  # key, indent, formatstr, expected
        (
            "Gumby",
            None,
            "{}{},",
            ["", "", "Gumby,"],
        ),  # key, indent, formatstr, expected
    )
    for key, indent, formatstr, expected in data:
        result = []
        epMunch.printkey(key, indent, formatstr, func=result.append)
        assert result == expected


def test_printkey_withprint(capsys):
    """py.est for printkey when func=None"""
    key, indent, formatstr, expected = ("Gumby", None, None, "\n\nGumby\n")
    epMunch.printkey(key, indent, formatstr)
    captured = capsys.readouterr()
    assert captured.out == expected


def test_printmunch():
    """py.test for printmunch"""
    data = (
        (
            dict(a=dict(aa=dict(z=-1, y=-2))),
            0,
            None,
            [
                "",
                "a                                                !-  EP_KEY         # use .eppykey",
                "            aa                                   !-  EPJOBJECT_NAME # use .eppyname",
                "            -1                                   !-  z",
                "            -2                                   !-  y",
            ],
        ),  # dct, indent, index, expected
        (
            dict(a=dict(aa=dict(z=-1, y=[dict(zz=1), dict(zz=2)]))),
            0,
            None,
            [
                "",
                "a                                                !-  EP_KEY         # use .eppykey",
                "            aa                                   !-  EPJOBJECT_NAME # use .eppyname",
                "            -1                                   !-  z",
                "                                                 !-  y",
                "                1                                    !-  zz #1",
                "                2                                    !-  zz #2",
            ],
        ),  # dct, indent, index, expected
        (
            dict(a=dict(aa=dict(z=-1, y=-2))),
            0,
            3,
            [
                "",
                "a                                                !-  EP_KEY         # use .eppykey",
                "            aa                                   !-  EPJOBJECT_NAME # use .eppyname",
                "            -1                                   !-  z #3",
                "            -2                                   !-  y #3",
            ],
        ),  # dct, indent, index, expected
        (
            dict(a=dict(aa=dict(z=-1, y=-2))),
            0,
            None,
            [
                "",
                "a                                                !-  EP_KEY         # use .eppykey",
                "            aa                                   !-  EPJOBJECT_NAME # use .eppyname",
                "            -1                                   !-  z",
                "            -2                                   !-  y",
            ],
        ),  # dct, indent, index, expected
    )  # noqa: E501
    for dct, indent, index, expected in data:
        dctstr = json.dumps(dct)
        fhandle = StringIO(dctstr)
        amunch = readepj.readepjjson(fhandle)
        result = []
        epMunch.printmunch(amunch, indent, index, result.append)
        assert result == expected


def test_printmunch_withprint(capsys):
    """py.test for printmunch with print"""
    (dct, indent, index, expected) = (
        dict(a=dict(aa=dict(z=-1, y=-2))),
        0,
        None,
        "\n".join(
            [
                "",
                "a                                                !-  EP_KEY         # use .eppykey",
                "            aa                                   !-  EPJOBJECT_NAME # use .eppyname",  # noqa: E501
                "            -1                                   !-  z",
                "            -2                                   !-  y",
            ]
        ),
    )
    dctstr = json.dumps(dct)
    fhandle = StringIO(dctstr)
    amunch = readepj.readepjjson(fhandle)
    epMunch.printmunch(amunch, indent, index)
    captured = capsys.readouterr()
    assert captured.out == expected + "\n"


def test_printmunch_ofEPSchema():
    """py.test for printmunch id you try to print EPSchemaMunch"""
    amunch = EPSchemaMunch(dict(a=1))
    expected = []
    result = []
    epMunch.printmunch(amunch)
    assert result == expected


class TestEPMunch_simple(object):
    """py.test for EPMunch"""

    def setup(self):
        dct = dict(a=dict(aa=dict(z=-1, y=-2)))
        dctstr = json.dumps(dct)
        fhandle = StringIO(dctstr)
        self.amunch = readepj.readepjjson(fhandle)

    def test_repr(self):
        """py.test for EPMunch.__repr__"""
        lst = [
            "",
            "a                                                !-  EP_KEY         # use .eppykey",
            "            aa                                   !-  EPJOBJECT_NAME # use .eppyname",  # noqa: E501
            "            -1                                   !-  z",
            "            -2                                   !-  y",
        ]
        expected = "\n".join(lst)
        result = self.amunch.__repr__()
        assert result == expected

    def test_str(self):
        """py.test for EPMunch.__str__"""
        lst = [
            "",
            "a                                                !-  EP_KEY         # use .eppykey",
            # should self.epobjects be updated here
            "            aa                                   !-  EPJOBJECT_NAME # use .eppyname",  # noqa: E501
            "            -1                                   !-  z",
            "            -2                                   !-  y",
        ]
        expected = "\n".join(lst)
        result = self.amunch.__str__()
        assert result == expected


@pytest.fixture
def simplemunch(request):
    """simple epmuch for tests"""
    dct = dict(a=dict(aa=dict(z=-1, y=-2)))
    dctstr = json.dumps(dct)
    fhandle = StringIO(dctstr)
    amunch = readepj.readepjjson(fhandle)
    # print('\n-----------------')
    # print('fixturename : %s' % request.fixturename)
    # print('scope       : %s' % request.scope)
    # print('function    : %s' % request.function.__name__)
    # print('cls         : %s' % request.cls)
    # print('module      : %s' % request.module.__name__)
    # print('fspath      : %s' % request.fspath)
    # print('-----------------')

    request.cls.amunch = amunch
    # see this link for more on how request works
    # https://docs.pytest.org/en/stable/fixture.html#request-context
    # still not clear to me right now
    #
    # You can pick up amunch in the test as self.amunch

    # print('\n-----------------')
    # print('fixturename : %s' % request.fixturename)
    # print('scope       : %s' % request.scope)
    # print('function    : %s' % request.function.__name__)
    # print('cls         : %s' % request.cls)
    # print('module      : %s' % request.module.__name__)
    # print('fspath      : %s' % request.fspath)
    # print('-----------------')
    yield


@pytest.mark.usefixtures("simplemunch")
class TestEPMunch_simple1(object):
    """py.test for EPMunch"""

    # add an eppy_ field
    # change and eppy_field
    # test to see if it works
    @pytest.mark.parametrize(
        "fname, fvalue, expected", [("eppy_field", 52, 52),],  # fname, fvalue, expected
    )
    def test_add_eppy_field(self, fname, fvalue, expected):
        """test adding a field that starts with 'eppy'"""
        # start with epMunch objects that are not epobjects
        # add a key, value
        self.amunch[fname] = fvalue
        assert self.amunch[fname] == expected
        self.amunch.eppy_hardcoded = fvalue
        assert self.amunch["eppy_hardcoded"] == expected
        # change a value
        newvalue = "new value"
        self.amunch[fname] = newvalue
        assert self.amunch[fname] == newvalue  # since it is not an epobject
        # it should change
        self.amunch.eppy_hardcoded = newvalue
        assert self.amunch.eppy_hardcoded == newvalue  # value should not change
        #
        # Now test epMunch objects that are epobjects
        # add a key, value
        #
        self.amunch.a.aa[fname] = fvalue
        assert self.amunch.a.aa[fname] == expected
        self.amunch.a.aa.eppy_hardcoded = fvalue
        assert self.amunch.a.aa["eppy_hardcoded"] == expected
        # change a value
        newvalue = "new value"
        self.amunch.a.aa[fname] = newvalue
        assert self.amunch.a.aa[fname] == expected  # since it is an epobject
        # it should not change
        self.amunch.a.aa.eppy_hardcoded = newvalue
        assert self.amunch.a.aa.eppy_hardcoded == expected  # since it is an epobject
        # it should not change


def test_change_eppyname():
    """pytest for changing eppyname"""
    # make the epmunch
    dct = dict(a=dict(aa=dict(z=-1, y=-2), bb=dict(z=1, y=2)))
    dctstr = json.dumps(dct)
    fhandle = StringIO(dctstr)
    amunch = readepj.readepjjson(fhandle)
    aa = amunch.a.aa
    aa["eppyname"] = "cc"
    # aa.eppyname = 'cc'
    assert aa == amunch.a.cc
    cc = amunch.a.cc
    # cc['eppyname'] = 'dd'
    cc.eppyname = "dd"
    assert cc == amunch.a.dd
    amunch.a.dd.z = 55
    dct = dict(a=dict(dd=dict(z=55, y=-2), bb=dict(z=1, y=2)))
    dctstr = json.dumps(dct)
    fhandle = StringIO(dctstr)
    bmunch = readepj.readepjjson(fhandle)
    # assert amunch == bmunch

    # remove eppykeys beflore comparison.
    # otherwise it go into a recursive loop because of 'eppy_epobjects'
    amunchd = amunch.toDict()
    amunchd = Munch.fromDict(amunchd)
    removeeppykeys(amunchd)
    bmunchd = bmunch.toDict()
    bmunchd = Munch.fromDict(bmunchd)
    removeeppykeys(bmunchd)

    assert amunchd == bmunchd


def test_delete():
    """py.test for EPMunch.delete"""
    # make the epmunch
    dct = dict(a=dict(aa=dict(z=-1, y=-2), bb=dict(z=1, y=2)))
    dctstr = json.dumps(dct)
    fhandle = StringIO(dctstr)
    amunch = readepj.readepjjson(fhandle)
    amunch.a.aa.delete()
    assert list(amunch.a.bb.eppy_epobjects.keys()) == ["bb"]
    # test when it not an epobject
    with pytest.raises(epMunch.NotEPObject):
        amunch.a.delete()
    amunch["a"]["bb"].delete()  # test when bb is not an attribute
    assert list(amunch.a.keys()) == []



# Copyright (c) 2020 Santosh Philip
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================

"""Experimental functions sit here
these functions will be unstable and the API's can change.
But we can experiment with them to see if the design is good"""



# Copyright (c) 2019 Santosh Philip
# =======================================================================
#  Distributed under the MIT License.
#  (See accompanying file LICENSE or copy at
#  http://opensource.org/licenses/MIT)
# =======================================================================
"""read and save idf files without using eppy"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


def removecomment(astr, cphrase):
    """
    the comment is similar to that in python.
    any charachter after the # is treated as a comment
    until the end of the line
    astr is the string to be de-commented
    cphrase is the comment phrase"""
    # linesep = mylib3.getlinesep(astr)
    alist = astr.splitlines()
    for i in range(len(alist)):
        alist1 = alist[i].split(cphrase)
        alist[i] = alist1[0]

    # return string.join(alist, linesep)
    return "\n".join(alist)


def readrawidf(fhandle):
    """read the idf file as a dict. Dict keys are idfobjkeys,
    dict values are list in list"""
    astr = fhandle.read()
    nocom = removecomment(astr, "!")
    idfst = nocom
    alist = idfst.split(";")
    rawdata = {}
    for element in alist:
        lst = element.split(",")
        lst = [item.strip() for item in lst]
        # key = lst[0].strip().upper()
        # not sure of implications here
        key = lst[0].strip()
        if not key:
            continue
        rawdata.setdefault(key, [])
        rawdata[key].append(lst)
    return rawdata


def rawidf2str(rawdata, order=None):  # rname var rawidf -> potential nameclash
    """string rep of rawdata"""
    lst = []
    if order:
        keys = order
    else:
        keys = rawdata.keys()
    for key in keys:
        try:
            for vals in rawdata[key]:
                lst.append("{},".format(key))
                for val in vals[1:]:
                    lst.append("     {},".format(val))
                lst.pop(-1)
                lst.append("    {};".format(val))
                lst.append("")
        except KeyError as e:
            continue
    return "\n".join(lst)



# Copyright (c) 2019-2021 Santosh Philip
# =======================================================================
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
# =======================================================================
"""py.test for idfjsonconverter"""

from io import StringIO

from eppy3000 import idfjsonconverter
from eppy3000 import installlocation
from tests import schemafortesting

import pytest

SCHEMA_FILE = schemafortesting.schema_file


class IDFtxt:
    """an IDF txt"""

    idftxt = """
  Version,9.1;

  ZoneHVAC:EquipmentList,
    SPACE1-1 Eq,             !- Name
    SequentialLoad,          !- Load Distribution Scheme
    ZoneHVAC:AirDistributionUnit,  !- Zone Equipment 1 Object Type
    SPACE1-1 ATU,            !- Zone Equipment 1 Name
    1,                       !- Zone Equipment 1 Cooling Sequence
    1;                       !- Zone Equipment 1 Heating or No-Load Sequence


NodeList,
    SPACE1-1 In Nodes,       !- Name
    SPACE1-1 In Node;        !- Node 1 Name

NodeList,
    Supply Air Temp Nodes 1, !- Name
    Main Cooling Coil 1 Inlet Node,  !- Node 1 Name
    Heat Recovery Supply Outlet,  !- Node 2 Name
    Main Heating Coil 1 Inlet Node,  !- Node 3 Name
    Reheat Coil Air Inlet Node,  !- Node 4 Name
    VAV Sys 1 Outlet Node;   !- Node 5 Name

BuildingSurface:Detailed,
    WALL-1PF,                !- Name
    WALL,                    !- Surface Type
    WALL-1,                  !- Construction Name
    PLENUM-1,                !- Zone Name
    Outdoors,                !- Outside Boundary Condition
    ,                        !- Outside Boundary Condition Object
    SunExposed,              !- Sun Exposure
    WindExposed,             !- Wind Exposure
    0.50000,                 !- View Factor to Ground
    4,                       !- Number of Vertices
    0.0,0.0,3.0,  !- X,Y,Z ==> Vertex 1 {m}
    0.0,0.0,2.4,  !- X,Y,Z ==> Vertex 2 {m}
    30.5,0.0,2.4,  !- X,Y,Z ==> Vertex 3 {m}
    30.5,0.0,3.0;  !- X,Y,Z ==> Vertex 4 {m}

Timestep,4;

Building,
    Bldg one,                !- Name
    30.,                     !- North Axis {deg}
    City,                    !- Terrain
    0.04,                    !- Loads Convergence Tolerance Value
    0.4,                     !- Temperature Convergence Tolerance Value {deltaC}
    FullExterior,            !- Solar Distribution
    25,                      !- Maximum Number of Warmup Days
    6;                       !- Minimum Number of Warmup Days

Zone,
    Main Zone,               !- Name
    0,                       !- Direction of Relative North {deg}
    0,                       !- X Origin {m}
    0,                       !- Y Origin {m}
    0,                       !- Z Origin {m}
    1,                       !- Type
    1,                       !- Multiplier
    autocalculate,           !- Ceiling Height {m}
    autocalculate;           !- Volume {m3}

Zone,
    Another Zone,               !- Name
    90,                       !- Direction of Relative North {deg}
    11,                       !- X Origin {m}
    22,                       !- Y Origin {m}
    33,                       !- Z Origin {m}
    1,                       !- Type
    1,                       !- Multiplier
    autocalculate,           !- Ceiling Height {m}
    autocalculate;           !- Volume {m3}

OutdoorAir:NodeList,
    OutsideAirInletNodes;    !- Node or NodeList Name 1

"""


def test_keymapping():
    """py.test for keymapping"""
    data = (
        (
            ("Gumby", "Softy"),
            ("gumby", "so", "softy", "Kamby"),
            {"Gumby": "gumby", "Softy": "softy"},
        ),
        # somekeys, allkeys, expected
    )
    for somekeys, allkeys, expected in data:
        result = idfjsonconverter.keymapping(somekeys, allkeys)
        assert result == expected


def test_idf2json_json2idf():
    """py.test for idf2json and json2idf"""
    data = ((IDFtxt.idftxt,),)  # idftxt
    for (idftxt,) in data:
        # convert idf to json, then json to idf and lastly idf to json
        # compare the first json and last json
        idfhandle = StringIO(idftxt)
        epsjsonschema = schemafortesting.schema
        jsonresult1 = idfjsonconverter.idf2json(idfhandle, epsjsonschema)
        jsonhandle = StringIO(jsonresult1)
        epsjsonschema = schemafortesting.schema
        idfresult1 = idfjsonconverter.json2idf(jsonhandle, epsjsonschema)

        idfhandle = StringIO(idfresult1)
        epsjsonschema = schemafortesting.schema
        jsonresult2 = idfjsonconverter.idf2json(idfhandle, epsjsonschema)
        assert jsonresult1 == jsonresult2


@pytest.mark.parametrize(
    "idftxt, expected",
    [
        (
            """
    ! Floor Area:              232.25 m2
    ! Number of Stories:       1

    Version,9.1;

    Timestep,6;

    Building,
        Bldg,                    !- Name
        0.0,                     !- North Axis {deg}
        Suburbs,                 !- Terrain
        0.05,                    !- Loads Convergence Tolerance Value
        0.05,                    !- Temperature Convergence Tolerance Value {deltaC}
        MinimalShadowing,        !- Solar Distribution
        30,                      !- Maximum Number of Warmup Days
        6;                       !- Minimum Number of Warmup Days
    """,
            "9.1",
        ),  # idftxt, expected
        (
            """
    !- Darwin Line endings 

    VERSION,
        9.0;                      !- Version Identifier

    """,
            "9.0",
        ),  # idftxt, expected
        (
            """
    !- Darwin Line endings 

    VERSION,
        9.0
        ;                      !- Version Identifier

    """,
            "9.0",
        ),  # idftxt, expected
        # --
        # single line version 1
        (
            """
    !- Darwin Line endings 

    Timestep,6;    VERSION,
      9.0
        ;                      !- Version Identifier

    """,
            "9.0",
        ),  # idftxt, expected
        # --
        # single line version 2
        (
            """
    !- Darwin Line endings 

    Timestep,6;    VERSION, 9.0
        ;                      !- Version Identifier

    """,
            "9.0",
        ),  # idftxt, expected
        # --
        # single line version 3
        (
            """
    !- Darwin Line endings 

    Timestep,6;    VERSION, 9.0 ;  !- Version Identifier

    """,
            "9.0",
        ),  # idftxt, expected
        # ---
        # no version in file
        (
            """
    !- Darwin Line endings 

    XERSION,
        9.0
        ;                      !- Version Identifier

    """,
            None,
        ),  # idftxt, expected
    ],
)
def test_getidfversion(idftxt, expected):
    """py.test for getidfversion"""
    fhandle = StringIO(idftxt)
    result = idfjsonconverter.getidfversion(fhandle)
    assert result == expected


@pytest.mark.parametrize(
    "idftxt, idffilename, epjfilename, epschemapath, expected",
    [
        # the first two commented out tests will work only if you installed EnergyPlus
        # (IDFtxt.idftxt, "aa.idf", "bb.epj", None, None), # idftxt, idffilename, epjfilename, epschemapath, expected
        # (IDFtxt.idftxt, "aa.idf", None, None, None), # idftxt, idffilename, epjfilename, epschemapath, expected
        (
            IDFtxt.idftxt,
            "aa.idf",
            "bb.epj",
            SCHEMA_FILE,
            None,
        ),  # idftxt, idffilename, epjfilename, epschemapath, expected
        (
            IDFtxt.idftxt,
            "aa.idf",
            None,
            SCHEMA_FILE,
            None,
        ),  # idftxt, idffilename, epjfilename, epschemapath, expected
    ],
)
def test_idffile2epjfile(
    tmp_path, idftxt, idffilename, epjfilename, epschemapath, expected
):
    """py.test for idffile2epjfile"""
    # 1. write idf to idffile
    # 2. convert idffile to epjfile -> the function we are testing
    # 3. convert idf to epj (expected)
    # 4. read epjfile and compare to epj in 3.
    # -
    # 1. write idf to idffile
    tmpdir = tmp_path
    tmpidffile = tmpdir / idffilename
    tmpidffile.write_text(idftxt)
    idffilepath = tmpidffile.resolve()
    # -
    if epjfilename:
        tmpepjfile = tmpdir / epjfilename
        epjfilepath = tmpepjfile.resolve()
    else:
        epjfilepath = None
        newepjname = f'{idffilename.split(".")[0]}.epj'
        tmpepjfile = tmpdir / newepjname
        newepjfilepath = tmpepjfile.resolve()
    # 2. convert idffile to epjfile -> the function we are testing
    savedepjpath = idfjsonconverter.idffile2epjfile(
        idffilepath, epjfilepath, epschemapath
    )
    # 3. convert idf to epj (expected)
    if epschemapath:
        new_epschemapath = epschemapath
    else:
        with open(idffilepath, "r") as idfhandle:
            version = idfjsonconverter.getidfversion(idfhandle)
        new_epschemapath = installlocation.schemapath(version)

    idfhandle = StringIO(idftxt)
    with open(new_epschemapath, "r") as epschemahandle:
        expected = idfjsonconverter.idf2json(idfhandle, epschemahandle)
    # 4. read epjfile and compare to epj in 3.
    result = open(savedepjpath, "r").read()
    assert result == expected


@pytest.mark.parametrize(
    "idftxt, epjname, idfname, schemapath, expected",
    [
        # the first two commented out tests will work only if you installed EnergyPlus
        # (IDFtxt.idftxt, "aa.epj", "bb.idf", None, None), # idftxt, epjname, idfname, schemapath, expected
        # (IDFtxt.idftxt, "aa.epj", None, None, None), # idftxt, epjname, idfname,schemapath, expected
        (
            IDFtxt.idftxt,
            "aa.epj",
            "bb.idf",
            SCHEMA_FILE,
            None,
        ),  # idftxt, epjname, idfname, schemapath, expected
        (
            IDFtxt.idftxt,
            "aa.epj",
            None,
            SCHEMA_FILE,
            None,
        ),  # idftxt, epjname,idfname, schemapath, expected
    ],
)
def test_epjfile2idffile(tmp_path, idftxt, epjname, idfname, schemapath, expected):
    """py.test for epjfile2idffile"""
    # 1. convert idf to epj
    # 2. write epj to epjfile
    # 3. convert epjfile to idffile -> the function we are testing
    # 4. read the spj file and comvert it to idf
    # 5. read idffile and compare to idf in 4.
    # -
    # 1. convert idf to epj
    idfhandle = StringIO(idftxt)
    if schemapath:
        new_epschemapath = schemapath
    else:
        version = idfjsonconverter.getidfversion(idfhandle)
        new_epschemapath = installlocation.schemapath(version)
    idfhandle = StringIO(idftxt)  # rest handle
    with open(new_epschemapath, "r") as schemahandle:
        epjstr = idfjsonconverter.idf2json(idfhandle, schemahandle)
    # 2. write epj to epjfile
    tmpdir = tmp_path
    tmpepjfile = tmpdir / epjname
    tmpepjfile.write_text(epjstr)
    epjfilepath = tmpepjfile.resolve()
    # -
    if idfname:
        tmpidffile = tmpdir / idfname
        idffilepath = tmpidffile.resolve()
    else:
        idffilepath = None
        newidfname = f'{epjname.split(".")[0]}.idf'
        tmpidffile = tmpdir / newidfname
        newidffilepath = tmpidffile.resolve()
    # 3. convert epjfile to idffile -> the function we are testing
    idfjsonconverter.epjfile2idffile(epjfilepath, idffilepath, schemapath)
    # 4. read the spj file and comvert it to idf
    with open(epjfilepath, "r") as epjhandle:
        expected = idfjsonconverter.json2idf(epjhandle, open(new_epschemapath, "r"))
    # 5. read idffile and compare to idf in 4.
    if idffilepath:
        result = open(idffilepath, "r").read()
    else:
        result = open(newidffilepath, "r").read()
    assert result == expected


@pytest.mark.parametrize(
    "lst, expected",
    [
        ([(1, 2), (3, 0), (4, "")], [(1, 2), (3, 0)]),  # lst, expected
        (
            [(1, 2), (33, ""), (3, 0), (4, ""), (5, "")],
            [(1, 2), (33, ""), (3, 0)],
        ),  # lst, expected
    ],
)
def test_removetrailingblanks(lst, expected):
    """py.test for removetrailingblanks"""
    result = idfjsonconverter.removetrailingblanks(lst)
    assert result == expected


@pytest.mark.parametrize(
    "idftxt, startfolder, endfolder, idfext, epjext, schemapath",
    [
        # the first  commented out test will work only if you installed EnergyPlus
        # (
        #     IDFtxt.idftxt,
        #     "sfolder",
        #     "efolder",
        #     "IDF",
        #     "epj",
        #     None,
        # ),  # idftxt, startfolder, endfolder, idfext, epjext, schemapath
        (
            IDFtxt.idftxt,
            "sfolder",
            "efolder",
            None,
            None,
            SCHEMA_FILE,
        ),  # idftxt, startfolder, endfolder, idfext, epjext, schemapath
        (
            IDFtxt.idftxt,
            "sfolder",
            "efolder",
            "IDF",
            "epj",
            SCHEMA_FILE,
        ),  # idftxt, startfolder, endfolder, idfext, epjext, schemapath
        (
            IDFtxt.idftxt,
            "sfolder",
            None,
            None,
            None,
            SCHEMA_FILE,
        ),  # idftxt, startfolder, endfolder, idfext, epjext, schemapath
    ],
)
def test_idffolder2epjfolder(
    tmp_path, idftxt, startfolder, endfolder, idfext, epjext, schemapath
):
    """py.test for idffolder2epjfolder"""
    # 1. set up a start folder
    # 2. use idftxt and put 2 idftxtfiles in that folder
    # 3. set up an end folder
    # 4. run conversion function idf to epj (startfolder to endfolder)
    # 5. do a in memory conversion for idfxtx to epj and back to idf
    # 6. compare contents of endfolder to idf in 5.

    # alternate methods
    # conversions
    # idffolder -> epjfolder -> idf1folder -> epj1folder -> idf2folder
    # idf1folder and idf2 folder should be identical

    # set endfolder
    if not endfolder:
        endfolder4test = startfolder
    else:
        endfolder4test = endfolder

    # setup extennsion values for use in the test
    if not idfext:
        idfext4test = "idf"
    else:
        idfext4test = idfext
    if not epjext:
        epjext4test = "epJSON"
    else:
        epjext4test = epjext

    # setup schemapath for use in test
    if schemapath:
        schemapath4test = schemapath
    else:
        version = idfjsonconverter.getidfversion(StringIO(idftxt))
        schemapath4test = installlocation.schemapath(version)

    # 1. set up a start folder
    tmpdir = tmp_path
    startfolder = tmpdir / startfolder
    startfolder.mkdir()
    # 2. use idftxt and put 2 idftxtfiles in that folder
    fnames = ["file1.idf", "file2.idf"]
    fnames = [f"file1.{idfext4test}", f"file2.{idfext4test}"]
    for fname in fnames:
        fpath = startfolder / fname
        fpath.write_text(idftxt)
    # 3. set up an end folder
    endfolder4test = tmpdir / endfolder4test
    try:
        endfolder4test.mkdir()
        passendfolder = endfolder4test
    except FileExistsError as e:
        passendfolder = endfolder
        # we are using the start folder as endfolder
    # 4. run conversion function idf to epj (startfolder to endfolder)
    idfjsonconverter.idffolder2epjfolder(
        startfolder,
        epjfolder=passendfolder,
        idfext=idfext,
        epjext=epjext,
        schemapath=schemapath,
    )
    # 5. do a in memory conversion for idftxt to epj
    expected = idfjsonconverter.idf2json(StringIO(idftxt), open(schemapath4test, "r"))
    # 6. compare contents of endfolder to idf in 5.
    for fname in fnames:
        fpath = endfolder4test / fname
        epjpath = fpath.with_suffix(f".{epjext4test}")
        result = open(epjpath, "r").read()
        assert result == expected


@pytest.mark.parametrize(
    "filelist, expectedfiles, ext",
    [
        (
            ["a.txt", "b.txt", "c.rtf"],
            ["a.txt", "b.txt"],
            "txt",
        ),  # filelist, expectedfiles, ext
        (["a.txt", "b.txt", "c.rtf"], [], "trt"),  # filelist, expectedfiles, ext
        (["a.txt", "b.txt", "c.rtf"], ["c.rtf"], "rtf"),  # filelist, expectedfiles, ext
    ],
)
def test_getfilepaths(tmp_path, filelist, expectedfiles, ext):
    """py.test getfilepaths"""
    # 0. setup variables
    # 1. make folder
    # 2. make subfolder
    # 3, make files in folder and subfolder
    # 4. run function for folder and ext
    # 5. compare result to expected

    # 0. setup variables
    folder, subfolder = "folder", "subfolder"
    # 1. make folder
    tmpdir = tmp_path
    thefolder = tmpdir / folder
    thefolder.mkdir()
    # 2. make subfolder
    thesubfolder = thefolder / subfolder
    thesubfolder.mkdir()
    # 3, make files in folder and subfolder
    for fname in filelist:
        fpath = thefolder / fname
        fpath.write_text("some text")
    for fname in filelist:
        fpath = thesubfolder / fname
        fpath.write_text("some text")
    # 4. run function for folder and ext
    result = idfjsonconverter.getfilepaths(thefolder, ext)
    expected = [thefolder / fname for fname in expectedfiles]
    # 5. compare result to expected
    assert set(result) == set(expected)
